# @title new perm cca

import numpy as np
from permcca.inference import permutation_inference
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import VarianceThreshold
from sklearn.pipeline import Pipeline

X = np.load("../assets/X.npy")
Y = np.load("../assets/Y.npy")

x_preproc = Pipeline(
    [
        ("scale", StandardScaler(with_mean=True, with_std=False)),
        ("var", VarianceThreshold(threshold=0.0)),
        # ("pca", PCA(n_components=10)),
    ]
).fit_transform(X)

y_preproc = Pipeline(
    [
        ("scale", StandardScaler(with_mean=True, with_std=False)),
        ("var", VarianceThreshold(threshold=0.0)),
        # ("pca", PCA(n_components=10)),
    ]
).fit_transform(Y)

for i in range(5, 11):
    p = permutation_inference(x_preproc, y_preproc, i)
    print(p)



# TO BE DONE

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_array
from scipy.linalg import null_space, svd, pinv, sqrtm


class RemoveCovars(BaseEstimator, TransformerMixin):
    def __init__(self, x_covars: np.ndarray, y_covars: np.ndarray):
        self.x_covars = x_covars
        self.y_covars = y_covars
        raise NotImplementedError("This class is not implemented yet.")

    def fit(self, x: np.ndarray, y: np.ndarray):
        x = check_array(x)
        y = check_array(y)
        return self

    def transform(self, x, y=None):
        x = check_array(x)
        y = check_array(y)
        return x - self.covars, y - self.covars

    def fit_transform(self, x, y=None, **fit_params):
        return self.fit(x, y).transform(x, y)


class SemiOrthogonal(BaseEstimator, TransformerMixin):
    """
    Compute a semi-orthogonal matrix according to
    the Huh-Jhun or Theil methods.

    Parameters:
    Z (ndarray): Input matrix of shape (N, R).
    Sel (ndarray or int or None): Selection matrix or selection method.
        If Sel is None, use the Huh-Jhun method.
        If Sel is an integer -1, use the Theil method with random selection.
        If Sel is a vector of indices or a boolean mask, use the Theil method with explicit selection.
        If Sel is a matrix, use the Theil method with the given selection matrix.

    Returns:
    Q (ndarray): Semi-orthogonal matrix of shape (N, N).
    """

    def __init__(self, selection=None):
        self.selection = selection

    def fit(self, X, y=None):
        # prepare the selection matrix
        if self.selection is None:
            pass
        else:
            # Theil method
            N, R = X.shape
            if isinstance(self.selection, np.ndarray):
                # Sel is a matrix
                return self
            elif np.isscalar(self.selection) and self.selection == -1:
                # Sel is -1
                rZ = np.linalg.matrix_rank(X)
                if rZ < R:
                    raise ValueError(
                        "Impossible to use the Theil method with this set of nuisance variables"
                    )
                _, iU, _ = np.unique(X, axis=0, return_index=True)
                np.random.shuffle(iU)
                unSel0 = []
                rnk0 = 0
                for u in iU:
                    unSel = np.concatenate((unSel0, [u]))
                    Zout = X[unSel, :]
                    rnk = np.linalg.matrix_rank(Zout)
                    if rnk > rnk0:
                        unSel0 = unSel
                        rnk0 = rnk
                    if rnk == R:
                        break
                selection = np.setdiff1d(np.arange(N), unSel)
                self.selection = np.eye(N)[:, selection]
            else:
                # Sel is a vector of indices
                if np.issubdtype(self.selection.dtype, np.bool_):
                    selection = np.nonzero(self.selection)[0]
                else:
                    selection = self.selection
                if selection[0] > 0:
                    unSel = np.setdiff1d(np.arange(N), selection)
                    if np.linalg.matrix_rank(X[unSel, :]) < R:
                        raise ValueError("Selected rows of nuisance not full rank")
                else:
                    raise ValueError("Invalid value for Sel")
                self.selection = np.eye(N)[:, selection]
        return self

    def transform(self, X, y=None):
        N, R = X.shape

        if self.selection is None:
            # Huh-Jhun method
            Q, D, _ = svd(null_space(X.T), full_matrices=False)
            Q = Q @ D
        else:
            Rz = np.eye(N) - X @ pinv(X)
            Q = (
                    Rz
                    @ self.selection
                    @ sqrtm(pinv(self.selection.T @ Rz @ self.selection))
            )
        return Q

    def fit_transform(self, X, y=None, **fit_params):
        return self.fit(X, y).transform(X, y)


def semiortho(Z, selection=None):

    if selection is None:
        # Huh-Jhun method
        Q, D, _ = svd(null_space(Z.T), full_matrices=False)
        Q = Q @ D
    else:
        # Theil method
        N, R = Z.shape
        if isinstance(selection, np.ndarray):
            # Sel is a matrix
            S = selection
        elif np.isscalar(selection) and selection == -1:
            # Sel is -1
            rZ = np.linalg.matrix_rank(Z)
            if rZ < R:
                raise ValueError(
                    "Impossible to use the Theil method with this set of nuisance variables"
                )
            _, iU, _ = np.unique(Z, axis=0, return_index=True)
            np.random.shuffle(iU)
            unSel0 = []
            rnk0 = 0
            for u in iU:
                unSel = np.concatenate((unSel0, [u]))
                Zout = Z[unSel, :]
                rnk = np.linalg.matrix_rank(Zout)
                if rnk > rnk0:
                    unSel0 = unSel
                    rnk0 = rnk
                if rnk == R:
                    break
            selection = np.setdiff1d(np.arange(N), unSel)
            S = np.eye(N)[:, selection]
        else:
            # Sel is a vector of indices
            if np.issubdtype(selection.dtype, np.bool_):
                selection = np.nonzero(selection)[0]
            if selection[0] > 0:
                unSel = np.setdiff1d(np.arange(N), selection)
                if np.linalg.matrix_rank(Z[unSel, :]) < R:
                    raise ValueError("Selected rows of nuisance not full rank")
            else:
                raise ValueError("Invalid value for Sel")
            S = np.eye(N)[:, selection]
        Rz = np.eye(N) - Z @ pinv(Z)
        Q = Rz @ S @ sqrtm(pinv(S.T @ Rz @ S))
    return Q


def center(X):
    """
    Mean center a matrix and remove constant columns.

    Parameters:
    X (ndarray): Matrix of shape (N, p) to center.

    Returns:
    X (ndarray): Centered matrix of shape (N, p') with p' <= p.
    """
    # icte = np.sum(np.diff(X, axis=0) ** 2, axis=0) == 0 # the original code
    icte = np.std(X, axis=0) != 0  # the inconstant columns

    X = X - np.mean(X, axis=0)
    X = X[:, icte]
    return X


def remove_covars(X, Y, covars_x=None, covars_y=None, partial=True, selection=None):
    identity = np.eye(X.shape[0])  # Create an n_samples x n_samples identity matrix
    if selection is None:
        selection = []
    X = center(X)
    Y = center(Y)
    if partial == True or partial == "bipartial":
        if covars_x is None:
            raise ValueError("covars_x must be specified for bi-partial CCA.")
        covars = center(covars_x)
        q_covars_x = semiortho(covars, selection)
        q_covars_y = q_covars_x.copy()

    elif partial == "semi":
        if covars_x is None and covars_y is None:
            raise ValueError(
                "Either covars_x or covars_y must be specified in semi-partial CCA."
            )
        covars_x = center(covars_x)
        covars_y = center(covars_y)
        q_covars_x = semiortho(covars_x, selection)
        q_covars_y = semiortho(covars_y, selection)
    else:
        q_covars_x = identity
        q_covars_y = identity
    X = q_covars_x.T @ X
    Y = q_covars_y.T @ Y
    n_samples_x, feat_num_x = X.shape
    n_samples_y, feat_num_y = Y.shape
    cova_num_x = covars_x.shape[1] if covars_x is not None else 0
    cova_num_y = covars_y.shape[1] if covars_y is not None else 0



import numpy as np
from sklearn.cross_decomposition import CCA
from scipy.linalg import null_space
from scipy.stats import pearsonr
from permcca.permute import PermutationMultiVar
import pingouin as pg


def can_corr(X_c, Y_c):
    return np.array([pearsonr(X_c[:, i], Y_c[:, i])[0] for i in range(X_c.shape[1])])


def _wilk_statistic(can_corr: np.ndarray):
    return -np.flip(np.cumsum(np.flip(np.log(1 - can_corr**2))))[0]


def _calc_cum_rs(X_c, Y_c, n_comps):
    latent_cum_rs = []
    for k in range(n_comps):
        cca = CCA(n_components=n_comps - k, scale=False)
        tmp_x, tmp_y = cca.fit_transform(X_c[:, k:], Y_c[:, k:])
        permed_rs = can_corr(tmp_x, tmp_y)
        cum_r = _wilk_statistic(permed_rs)
        latent_cum_rs.append(cum_r)

    return np.array(latent_cum_rs)


def permutation_inference(X, Y, n_comps=None, n_perms=1000, fwe_correction=True):
    # TODO: add support for passing model
    # TODO: add support for passing n_jobs
    if X.shape[0] != Y.shape[0]:
        raise ValueError("Y and X do not have same number of rows.")
    if n_comps is None:
        n_comps = min(X.shape[1], Y.shape[1])
    cca = CCA(n_components=n_comps, scale=False)
    cca.fit_transform(X, Y)
    X_c = X @ np.hstack((cca.x_weights_, null_space(cca.x_weights_.T)))
    Y_c = Y @ np.hstack((cca.y_weights_, null_space(cca.y_weights_.T)))
    cnt = np.zeros(n_comps)
    # For each permutation
    seeds = np.random.randint(0, 10000, size=n_perms)
    latent_cum_rs = _calc_cum_rs(X_c, Y_c, n_comps)
    for perm in range(0, n_perms):
        perm_X_c, perm_Y_c = PermutationMultiVar(
            random_state=seeds[perm], shuffle_target=True, selection=None
        ).fit_transform(X_c, Y_c)
        permuted_latent_cum_rs = _calc_cum_rs(perm_X_c, perm_Y_c, n_comps)
        cnt += (permuted_latent_cum_rs >= latent_cum_rs).astype(int)

    # Compute p-values
    punc = np.minimum(cnt / n_perms, 1)
    if fwe_correction:
        p_fwe = pg.multicomp(punc, method="fdr_bh")[1]
        return p_fwe
    else:
        return punc


"""
from concurrent.futures import ThreadPoolExecutor

def permute_and_calc_cum_rs(seed, X_c, Y_c, n_comps):
    perm_X_c, perm_Y_c = PermutationMultiVar(
        random_state=seed, shuffle_target=True, selection=None
    ).fit_transform(X_c, Y_c)
    return _calc_cum_rs(perm_X_c, perm_Y_c, n_comps)

def permutation_inference(X, Y, n_comps=None, n_perms=1000, n_threads=4):
    if X.shape[0] != Y.shape[0]:
        raise ValueError("Y and X do not have same number of rows.")
    if n_comps is None:
        n_comps = min(X.shape[1], Y.shape[1])
    cca = CCA(n_components=n_comps, scale=False)
    cca.fit_transform(X, Y)
    X_c = X @ np.hstack((cca.x_weights_, null_space(cca.x_weights_.T)))
    Y_c = Y @ np.hstack((cca.y_weights_, null_space(cca.y_weights_.T)))
    cnt = np.zeros(n_comps)
    seeds = np.random.randint(0, 10000, size=n_perms)
    latent_cum_rs = _calc_cum_rs(X_c, Y_c, n_comps)
    print(latent_cum_rs)

    with ThreadPoolExecutor(max_workers=n_threads) as executor:
        results = list(executor.map(permute_and_calc_cum_rs, seeds, [X_c]*n_perms, [Y_c]*n_perms, [n_comps]*n_perms))
    
    for permuted_latent_cum_rs in results:
        cnt += (permuted_latent_cum_rs >= latent_cum_rs).astype(int)

    punc = np.minimum(cnt / n_perms, 1)
    return punc
    """



"""
Make sure these models are using sklearn's API.
This scripts provides a basic cca models, but the further inference scripts should be compatible with
other cca models, such as models from cca-zoo packages
"""

# Add a model that is exactly same as the matlab one

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_array


class ClassicCCA(BaseEstimator, TransformerMixin):
    def __init__(self, n_components=2):
        self.n_components = n_components
        self.r_ = None
        self.a_ = None
        self.b_ = None

    def fit(self, X, Y):
        X = check_array(X)
        Y = check_array(Y)
        K = self.n_components
        # Use the QR decomposition methods
        # In general your sample size should be at 5-10 times the number of variables, so should use reduced mode
        Qx, Rx = np.linalg.qr(X, mode="reduced")
        Qy, Ry = np.linalg.qr(X, mode="reduced")

        # compute SVD on Qx.T.dot(Qy)
        U, S, Vh = np.linalg.svd(Qx.T.dot(Qy))
        V = Vh.T
        # A = np.linalg.inv(Rx).dot(U)
        # B = np.linalg.inv(Ry).dot(V)
        if K >= min(X.shape[1], Y.shape[1]):
            raise ValueError("K is too large")
        elif K < 1:
            raise ValueError("K should be positive")
        raise NotImplementedError(
            "Need to implement the rest of the code, It;s still wrong!!"
        )
        self.a_ = np.linalg.inv(Rx).dot(U[:, :K])
        self.b_ = np.linalg.inv(Ry).dot(V[:, :K])

        return self

    def transform(self, X, Y=None):
        X = check_array(X)
        if Y is not None:
            Y = check_array(Y)
        print(Y)
        u = X.dot(self.a_)
        v = Y.dot(self.b_)

        return u, v


#%%
"""
function [A,B,r,U,V,stats] = canoncorr2(X,Y)
%CANONCORR Canonical correlation analysis.
%   [A,B] = CANONCORR(X,Y) computes the sample canonical coefficients for
%   the N-by-P1 and N-by-P2 data matrices X and Y.  X and Y must have the
%   same number of observations (rows) but can have different numbers of
%   variables (cols).  A and B are P1-by-D and P2-by-D matrices, where D =
%   min(rank(X),rank(Y)).  The jth columns of A and B contain the canonical
%   coefficients, i.e. the linear combination of variables making up the
%   jth canonical variable for X and Y, respectively.  Columns of A and B
%   are scaled to make COV(U) and COV(V) (see below) the identity matrix.
%   If X or Y are less than full rank, CANONCORR gives a warning and
%   returns zeros in the rows of A or B corresponding to dependent columns
%   of X or Y.
%
%   [A,B,R] = CANONCORR(X,Y) returns the 1-by-D vector R containing the
%   sample canonical correlations.  The jth element of R is the correlation
%   between the jth columns of U and V (see below).
%
%   [A,B,R,U,V] = CANONCORR(X,Y) returns the canonical variables, also
%   known as scores, in the N-by-D matrices U and V.  U and V are computed
%   as
%
%      U = (X - repmat(mean(X),N,1))*A and
%      V = (Y - repmat(mean(Y),N,1))*B.
%
%   [A,B,R,U,V,STATS] = CANONCORR(X,Y) returns a structure containing
%   information relating to the sequence of D null hypotheses H0_K, that
%   the (K+1)st through Dth correlations are all zero, for K = 0:(D-1).
%   STATS contains eight fields, each a 1-by-D vector with elements
%   corresponding to values of K:
%
%      Wilks:    Wilks' lambda (likelihood ratio) statistic
%      chisq:    Bartlett's approximate chi-squared statistic for H0_K,
%                with Lawley's modification
%      pChisq:   the right-tail significance level for CHISQ
%      F:        Rao's approximate F statistic for H0_K
%      pF:       the right-tail significance level for F
%      df1:      the degrees of freedom for the chi-squared statistic,
%                also the numerator degrees of freedom for the F statistic
%      df2:      the denominator degrees of freedom for the F statistic
%
%   Example:
%
%      load carbig;
%      X = [Displacement Horsepower Weight Acceleration MPG];
%      nans = sum(isnan(X),2) > 0;
%      [A B r U V] = canoncorr(X(~nans,1:3), X(~nans,4:5));
%
%      plot(U(:,1),V(:,1),'.');
%      xlabel('0.0025*Disp + 0.020*HP - 0.000025*Wgt');
%      ylabel('-0.17*Accel + -0.092*MPG')
%
%   See also PRINCOMP, MANOVA1.

%   References:
%     [1] Krzanowski, W.J., Principles of Multivariate Analysis,
%         Oxford University Press, Oxford, 1988.
%     [2] Seber, G.A.F., Multivariate Observations, Wiley, New York, 1984.

%   Copyright 1993-2009 The MathWorks, Inc.
%   $Revision: 1.1.8.1 $  $Date: 2010/03/16 00:12:35 $

if nargin < 2
    error('stats:canoncorr:TooFewInputs','Requires two arguments.');
end

[n,p1] = size(X);
if size(Y,1) ~= n
    error('stats:canoncorr:InputSizeMismatch',...
          'X and Y must have the same number of rows.');
elseif n == 1
    error('stats:canoncorr:NotEnoughData',...
          'X and Y must have more than one row.');
end
p2 = size(Y,2);


% Factor the inputs, and find a full rank set of columns if necessary
[Q1,T11,perm1] = qr(X,0);
rankX = sum(abs(diag(T11)) > eps(abs(T11(1)))*max(n,p1));
if rankX == 0
    error('stats:canoncorr:BadData',...
          'X must contain at least one non-constant column');
elseif rankX < p1
    warning('stats:canoncorr:NotFullRank','X is not full rank.');
    Q1 = Q1(:,1:rankX); T11 = T11(1:rankX,1:rankX);
end
[Q2,T22,perm2] = qr(Y,0);
rankY = sum(abs(diag(T22)) > eps(abs(T22(1)))*max(n,p2));
if rankY == 0
    error('stats:canoncorr:BadData',...
          'Y must contain at least one non-constant column');
elseif rankY < p2
    warning('stats:canoncorr:NotFullRank','Y is not full rank.');
    Q2 = Q2(:,1:rankY); T22 = T22(1:rankY,1:rankY);
end

% Compute canonical coefficients and canonical correlations.  For rankX >
% rankY, the economy-size version ignores the extra columns in L and rows
% in D. For rankX < rankY, need to ignore extra columns in M and D
% explicitly. Normalize A and B to give U and V unit variance.
d = min(rankX,rankY);
[L,D,M] = svd(Q1' * Q2,0);
A = T11 \ L(:,1:d) * sqrt(n-1);
B = T22 \ M(:,1:d) * sqrt(n-1);
r = min(max(diag(D(:,1:d))', 0), 1); % remove roundoff errs

% Put coefficients back to their full size and their correct order
A(perm1,:) = [A; zeros(p1-rankX,d)];
B(perm2,:) = [B; zeros(p2-rankY,d)];

% Compute the canonical variates
if nargout > 3
    U = X * A;
    V = Y * B;
end
"""
